import importlib
from typing import Never

import typer
from rich import print
from rich.panel import Panel

from engin import Engin


def print_error(msg: str) -> Never:
    print(
        Panel(
            title="Error",
            renderable=msg,
            title_align="left",
            border_style="red",
            highlight=True,
        )
    )
    raise typer.Exit(code=1)


COMMON_HELP = {
    "app": (
        "The import path of your Engin instance, in the form 'package:application'"
        ", e.g. 'app.main:engin'"
    )
}


def get_engin_instance(app: str) -> tuple[str, str, Engin]:
    try:
        module_name, engin_name = app.split(":", maxsplit=1)
    except ValueError:
        print_error("Expected an argument of the form 'module:attribute', e.g. 'myapp:engin'")

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        print_error(f"Unable to find module '{module_name}'")

    try:
        instance = getattr(module, engin_name)
    except AttributeError:
        print_error(f"Module '{module_name}' has no attribute '{engin_name}'")

    if not isinstance(instance, Engin):
        print_error(f"'{app}' is not an Engin instance")

    return module_name, engin_name, instance



import inspect
import logging
import typing
from collections.abc import Awaitable, Callable
from contextlib import AsyncExitStack
from dataclasses import dataclass
from enum import Enum
from types import TracebackType
from typing import TypeAlias

import anyio
from anyio import get_cancelled_exc_class
from exceptiongroup import BaseExceptionGroup, catch

from engin._shutdown import ShutdownSwitch

if typing.TYPE_CHECKING:
    from anyio.abc import TaskGroup

LOG = logging.getLogger("engin")

TaskFactory: TypeAlias = Callable[[], Awaitable[None]]


class OnException(Enum):
    SHUTDOWN = 0
    """
    Cancel all other supervised tasks and shutdown the Engin.
    """

    RETRY = 1
    """
    Retry the task.
    """

    IGNORE = 2
    """
    The task will be not be retried and the engin will not be stopped, other tasks will
    continue to run.
    """


@dataclass(kw_only=True, slots=True, eq=False)
class _SupervisorTask:
    """
    Attributes:
        - factory: a coroutine function that can create the task.
        - on_exception: determines the behaviour when task raises an exception.
        - complete: will be set to true if task stops for any reason except cancellation.
        - last_exception: the last exception raised by the task.
    """

    factory: TaskFactory
    on_exception: OnException
    complete: bool = False
    last_exception: Exception | None = None

    async def __call__(self) -> None:
        # loop to allow for restarting erroring tasks
        while True:
            try:
                await self.factory()
                self.complete = True
                return
            except get_cancelled_exc_class():
                LOG.info(f"{self.name} cancelled")
                raise
            except Exception as err:
                LOG.error(f"Supervisor: {self.name} raised {type(err).__name__}", exc_info=err)
                self.last_exception = err

                if self.on_exception == OnException.IGNORE:
                    self.complete = True
                    return

                if self.on_exception == OnException.RETRY:
                    continue

                if self.on_exception == OnException.SHUTDOWN:
                    self.complete = True
                    raise

    @property
    def name(self) -> str:
        factory = self.factory
        if inspect.ismethod(factory):
            return f"{factory.__self__.__class__.__name__}.{factory.__func__.__name__}"
        if inspect.isclass(factory):
            return type(factory).__name__
        if inspect.isfunction(factory):
            return factory.__name__
        return str(factory)


class Supervisor:
    def __init__(self, shutdown: ShutdownSwitch) -> None:
        self._tasks: list[_SupervisorTask] = []
        self._shutdown = shutdown
        self._is_complete: bool = False

        self._exit_stack: AsyncExitStack | None = None
        self._task_group: TaskGroup | None = None

    def supervise(
        self, func: TaskFactory, *, on_exception: OnException = OnException.SHUTDOWN
    ) -> None:
        self._tasks.append(_SupervisorTask(factory=func, on_exception=on_exception))

    @property
    def empty(self) -> bool:
        return not self._tasks

    async def __aenter__(self) -> None:
        if not self._tasks:
            return

        def _handler(_: BaseExceptionGroup) -> None:
            self._shutdown.set()

        self._exit_stack = AsyncExitStack()
        await self._exit_stack.__aenter__()
        self._exit_stack.enter_context(catch({Exception: _handler}))
        self._task_group = await self._exit_stack.enter_async_context(
            anyio.create_task_group()
        )

        for task in self._tasks:
            self._task_group.start_soon(task, name=task.name)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        if not self._tasks:
            return

        if self._exit_stack:
            await self._exit_stack.__aexit__(exc_type, exc_value, traceback)



import contextlib
import socketserver
import threading
from http.server import BaseHTTPRequestHandler
from pathlib import Path
from time import sleep
from typing import Annotated, Any

import typer
from rich import print

from engin import Entrypoint, Invoke, TypeId
from engin._cli._common import COMMON_HELP, get_engin_instance
from engin._dependency import Dependency, Provide, Supply
from engin.extensions.asgi import ASGIEngin

try:
    from engin.extensions.fastapi import APIRouteDependency
except ImportError:
    APIRouteDependency = None  # type: ignore[assignment,misc]

cli = typer.Typer()


_APP_ORIGIN = ""


@cli.command(name="graph")
def serve_graph(
    app: Annotated[
        str,
        typer.Argument(help=COMMON_HELP["app"]),
    ],
) -> None:
    """
    Creates a visualisation of your application's dependencies.
    """
    module_name, _, instance = get_engin_instance(app)

    global _APP_ORIGIN
    _APP_ORIGIN = module_name.split(".", maxsplit=1)[0]

    nodes = instance.graph()

    # transform dependencies into mermaid syntax
    dependencies = [
        f"{_render_node(node.parent)} --> {_render_node(node.node)}"
        for node in nodes
        if node.parent is not None
        and not (node.node.block_name and node.node.block_name == node.parent.block_name)
    ]

    blocks = {node.node.block_name for node in nodes if node.node.block_name is not None}

    # group blocks into subgraphs
    for block in blocks:
        dependencies.append(f"subgraph {block}")
        dependencies.extend(
            [
                f"{_render_node(node.parent, False)} --> {_render_node(node.node, False)}"
                for node in nodes
                if node.parent is not None
                and node.node.block_name == block
                and node.parent.block_name == block
            ]
        )
        dependencies.append("end")

    html = (
        _GRAPH_HTML.replace("%%DATA%%", "\n".join(dependencies))
        .replace(
            "%%LEGEND%%",
            ASGI_ENGIN_LEGEND if isinstance(instance, ASGIEngin) else DEFAULT_LEGEND,
        )
        .encode("utf8")
    )

    class Handler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            self.send_response(200, "OK")
            self.send_header("Content-type", "html")
            self.end_headers()
            self.wfile.write(html)

        def log_message(self, format: str, *args: Any) -> None:
            return

    def _start_server() -> None:
        with socketserver.TCPServer(("localhost", 8123), Handler) as httpd:
            print("Serving dependency graph on http://localhost:8123")
            httpd.serve_forever()

    server_thread = threading.Thread(target=_start_server)
    server_thread.daemon = True  # Daemonize the thread so it exits when the main script exits
    server_thread.start()

    with contextlib.suppress(KeyboardInterrupt):
        wait_for_interrupt()

    print("Exiting the server...")


def wait_for_interrupt() -> None:
    sleep(10000)


_BLOCK_IDX: dict[str, int] = {}
_SEEN_BLOCKS: list[str] = []


def _render_node(node: Dependency, render_block: bool = True) -> str:
    node_id = id(node)
    md = ""
    style = ""

    # format block name
    if render_block and (n := node.block_name):
        md += f"_{n}_\n"

    node_root_package = node.source_package.split(".", maxsplit=1)[0]
    if node_root_package != _APP_ORIGIN:
        if style:
            style += "E"
        else:
            style = "external"

    if style:
        style = f":::{style}"

    if isinstance(node, Supply):
        md += f"{_short_name(node.return_type_id)}"
        return f'{node_id}("`{md}`"){style}'
    if isinstance(node, Provide):
        md += f"{_short_name(node.return_type_id)}"
        return f'{node_id}["`{md}`"]{style}'
    if isinstance(node, Entrypoint):
        entrypoint_type = node.parameter_type_ids[0]
        md += f"{entrypoint_type}"
        return f'{node_id}[/"`{md}`"\\]{style}'
    if isinstance(node, Invoke):
        md += f"{node.func_name}"
        return f'{node_id}[/"`{md}`"/]{style}'
    if isinstance(node, APIRouteDependency):
        md += f"{node.name}"
        return f'{node_id}[["`{md}`"]]{style}'
    else:
        return f'{node_id}["`{node.name}`"]{style}'


def _short_name(name: TypeId) -> str:
    return str(name).rsplit(".", maxsplit=1)[-1]


_GRAPH_HTML = (Path(__file__).parent / "_graph.html").read_text()

DEFAULT_LEGEND = (
    "0[/Invoke/] ~~~ 1[/Entrypoint\\] ~~~ 2[Provide] ~~~ 3(Supply)"
    ' ~~~ 4["`Block Grouping`"]:::b0 ~~~ 5[External Dependency]:::external'
)
ASGI_ENGIN_LEGEND = DEFAULT_LEGEND + " ~~~ 6[[API Route]]"



import asyncio
from asyncio import TaskGroup

from engin import Engin
from engin._engin import _EnginState


async def test_engin_signal_handling():
    engin = Engin()

    async with TaskGroup() as tg:
        tg.create_task(engin.run())
        # give it time to startup
        await asyncio.sleep(0.1)
        assert engin._state == _EnginState.RUNNING
        await engin.stop()
        assert engin._state == _EnginState.SHUTDOWN



import asyncio
import logging
import os
import signal
from asyncio import Event, Task
from collections import defaultdict
from contextlib import AsyncExitStack
from enum import Enum
from itertools import chain
from types import FrameType
from typing import ClassVar

from anyio import CancelScope, create_task_group, get_cancelled_exc_class

from engin._assembler import AssembledDependency, Assembler
from engin._dependency import Invoke, Provide, Supply
from engin._graph import DependencyGrapher, Node
from engin._lifecycle import Lifecycle
from engin._option import Option
from engin._shutdown import ShutdownSwitch
from engin._supervisor import Supervisor
from engin._type_utils import TypeId
from engin.exceptions import EnginError

_OS_IS_WINDOWS = os.name == "nt"
LOG = logging.getLogger("engin")


class _EnginState(Enum):
    IDLE = 0
    """
    Not yet started.
    """

    STARTED = 1
    """
    Engin started via .start() call
    """

    RUNNING = 2
    """
    Engin running via .run() call
    """

    STOPPING = 3
    """
    Engin stopped via .stop() call
    """

    SHUTDOWN = 4
    """
    Engin has performed shutdown
    """


class Engin:
    """
    The Engin is a modular application defined by a collection of options.

    Users should instantiate the Engin with a number of options, where options can be an
    instance of Provide, Invoke, or a collection of these combined in a Block.

    To create a useful application, users should pass in one or more providers (Provide or
    Supply) and at least one invocation (Invoke or Entrypoint).

    When instantiated the Engin can be run. This is typically done via the `run` method,
    but for advanced usecases it can be easier to use the `start` and `stop` methods.

    When ran the Engin takes care of the complete application lifecycle:

    1. The Engin assembles all Invocations. Only Providers that are required to satisfy
       the Invoke options parameters are assembled.
    2. All Invocations are run sequentially in the order they were passed in to the Engin.
    3. Lifecycle Startup tasks registered by assembled dependencies are run sequentially.
    4. The Engin waits for a stop signal, i.e. SIGINT or SIGTERM, or for something to
       set the ShutdownSwitch event.
    5. Lifecyce Shutdown tasks are run in the reverse order to the Startup order.

    Examples:
        ```python
        import asyncio

        from httpx import AsyncClient

        from engin import Engin, Invoke, Lifecycle, Provide


        def httpx_client(lifecycle: Lifecycle) -> AsyncClient:
            client = AsyncClient()
            lifecycle.append(client)
            return client


        async def main(http_client: AsyncClient) -> None:
            print(await http_client.get("https://httpbin.org/get"))

        engin = Engin(Provide(httpx_client), Invoke(main))

        asyncio.run(engin.run())
        ```
    """

    _LIB_OPTIONS: ClassVar[list[Option]] = [Provide(Lifecycle), Provide(Supervisor)]

    def __init__(self, *options: Option) -> None:
        """
        Initialise the class with the provided options.

        Examples:
            >>> engin = Engin(Provide(construct_a), Invoke(do_b), Supply(C()), MyBlock())

        Args:
            *options: an instance of Provide, Supply, Invoke, Entrypoint or a Block.
        """
        self._state = _EnginState.IDLE
        self._stop_requested_event = ShutdownSwitch()
        self._stop_complete_event = Event()
        self._exit_stack: AsyncExitStack = AsyncExitStack()
        self._shutdown_task: Task | None = None
        self._run_task: Task | None = None
        self._assembler = Assembler([])

        self._providers: dict[TypeId, Provide] = {
            TypeId.from_type(Assembler): Supply(self._assembler),
            TypeId.from_type(ShutdownSwitch): Supply(self._stop_requested_event),
        }
        self._multiproviders: dict[TypeId, list[Provide]] = defaultdict(list)
        self._invocations: list[Invoke] = []

        # populates the above
        for option in chain(self._LIB_OPTIONS, options):
            option.apply(self)

        multi_providers = [p for multi in self._multiproviders.values() for p in multi]

        for provider in chain(self._providers.values(), multi_providers):
            self._assembler.add(provider)

    @property
    def assembler(self) -> Assembler:
        return self._assembler

    async def run(self) -> None:
        """
        Run the engin.

        The engin will run until it is stopped via an external signal (i.e. SIGTERM or
        SIGINT), the `stop` method is called on the engin, or a lifecycle task errors.
        """
        await self.start()

        # engin failed to start, so exit early
        if self._state != _EnginState.STARTED:
            return

        self._state = _EnginState.RUNNING

        async with create_task_group() as tg:
            tg.start_soon(_stop_engin_on_signal, self._stop_requested_event)
            try:
                await self._stop_requested_event.wait()
                await self._shutdown()
            except get_cancelled_exc_class():
                with CancelScope(shield=True):
                    await self._shutdown()

    async def start(self) -> None:
        """
        Start the engin.

        This is an alternative to calling `run`. This method waits for the startup
        lifecycle to complete and then returns. The caller is then responsible for
        calling `stop`.
        """
        if self._state != _EnginState.IDLE:
            raise EnginError("Engin is not idle, unable to start")

        LOG.info("starting engin")
        assembled_invocations: list[AssembledDependency] = [
            await self._assembler.assemble(invocation) for invocation in self._invocations
        ]

        for invocation in assembled_invocations:
            try:
                await invocation()
            except Exception as err:
                name = invocation.dependency.name
                LOG.error(f"invocation '{name}' errored, exiting", exc_info=err)
                return

        lifecycle = await self._assembler.build(Lifecycle)
        supervisor = await self._assembler.build(Supervisor)

        if not supervisor.empty:
            lifecycle.append(supervisor)

        try:
            for hook in lifecycle.list():
                await asyncio.wait_for(self._exit_stack.enter_async_context(hook), timeout=15)
        except Exception as err:
            if isinstance(err, TimeoutError):
                msg = "lifecycle startup task timed out after 15s, exiting"
            else:
                msg = "lifecycle startup task errored, exiting"
            LOG.error(msg, exc_info=err)
            await self._shutdown()
            return

        LOG.info("startup complete")
        self._state = _EnginState.STARTED

    async def stop(self) -> None:
        """
        Stop the engin.

        This method will wait for the shutdown lifecycle to complete before returning.
        Note this method can be safely called at any point, even before the engin is
        started.
        """
        # If the Engin was ran via `start()` perform shutdown directly
        if self._state == _EnginState.STARTED:
            await self._shutdown()
        # If the Engin was ran via `run()` notify via event
        elif self._state == _EnginState.RUNNING:
            self._stop_requested_event.set()
            await self._stop_complete_event.wait()

    def graph(self) -> list[Node]:
        grapher = DependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve(self._invocations)

    async def _shutdown(self) -> None:
        LOG.info("stopping engin")
        await self._exit_stack.aclose()
        self._stop_complete_event.set()
        LOG.info("shutdown complete")
        self._state = _EnginState.SHUTDOWN


async def _stop_engin_on_signal(stop_requested_event: Event) -> None:
    """
    A task that waits for a stop signal (SIGINT/SIGTERM) and notifies the given event.
    """
    try:
        # try to gracefully handle sigint/sigterm
        if not _OS_IS_WINDOWS:
            loop = asyncio.get_running_loop()
            for signame in (signal.SIGINT, signal.SIGTERM):
                loop.add_signal_handler(signame, stop_requested_event.set)

            await stop_requested_event.wait()
        else:
            should_stop = False

            # windows does not support signal_handlers, so this is the workaround
            def ctrlc_handler(sig: int, frame: FrameType | None) -> None:
                nonlocal should_stop
                if should_stop:
                    raise KeyboardInterrupt("Forced keyboard interrupt")
                should_stop = True

            signal.signal(signal.SIGINT, ctrlc_handler)

            while not should_stop:
                # In case engin is stopped via external `stop` call.
                if stop_requested_event.is_set():
                    return
                await asyncio.sleep(0.1)

            stop_requested_event.set()
    except asyncio.CancelledError:
        pass



import asyncio
from collections.abc import Iterable
from contextlib import asynccontextmanager
from datetime import datetime

import pytest

from engin import Engin, Entrypoint, Invoke, Lifecycle, Provide
from engin.exceptions import ProviderError
from tests.deps import ABlock


class A:
    def __init__(self): ...


class B:
    def __init__(self): ...


class C:
    def __init__(self): ...


async def test_engin():
    def a() -> A:
        return A()

    def b(_: A) -> B:
        return B()

    def c(_: B) -> C:
        return C()

    def multi_a() -> list[A]:
        return [A()]

    def multi_a_2() -> list[A]:
        return [A(), A()]

    def main(c: C, multi_a: list[A]) -> None:
        assert isinstance(c, C)
        assert len(multi_a) == 3

    engin = Engin(
        Provide(a), Provide(b), Provide(c), Provide(multi_a), Provide(multi_a_2), Invoke(main)
    )

    await engin.start()
    await engin.stop()


async def test_engin_with_block():
    def main(dt: datetime, floats: list[float]) -> None:
        assert isinstance(dt, datetime)
        assert isinstance(floats, list)
        assert all(isinstance(x, float) for x in floats)

    engin = Engin(ABlock(), Invoke(main))

    await engin.start()
    await engin.stop()


async def test_engin_error_handling():
    async def raise_value_error() -> int:
        raise ValueError("foo")

    async def main(foo: int) -> None:
        return

    engin = Engin(Provide(raise_value_error), Invoke(main))

    with pytest.raises(ProviderError, match="foo"):
        await engin.run()


async def test_engin_with_entrypoint():
    provider_called = False

    def a() -> A:
        nonlocal provider_called
        provider_called = True
        return A()

    engin = Engin(Provide(a), Entrypoint(A))

    await engin.start()
    await engin.stop()

    assert provider_called


async def test_engin_with_lifecycle():
    state = 0

    @asynccontextmanager
    async def lifespan_task() -> Iterable[None]:
        nonlocal state
        state = 1
        yield
        state = 2

    def foo(lifecycle: Lifecycle) -> None:
        lifecycle.append(lifespan_task())

    engin = Engin(Invoke(foo))

    await engin.start()
    assert state == 1

    await engin.stop()
    assert state == 2


async def test_engin_with_lifecycle_using_run():
    state = 0

    @asynccontextmanager
    async def lifespan_task() -> Iterable[None]:
        nonlocal state
        state = 1
        yield
        state = 2

    def foo(lifecycle: Lifecycle) -> None:
        lifecycle.append(lifespan_task())

    engin = Engin(Invoke(foo))

    async def _stop_task():
        await asyncio.sleep(0.25)
        # lifecycle should have started by now
        assert state == 1
        await engin.stop()

    await asyncio.gather(engin.run(), _stop_task())
    # lifecycle should have stopped by now
    assert state == 2


def test_engin_graph():
    def a() -> A:
        return A()

    def b(_: A) -> B:
        return B()

    def c(_: B) -> C:
        return C()

    def main(c: C) -> None:
        assert isinstance(c, C)

    engin = Engin(Provide(a), Provide(b), Provide(c), Invoke(main))

    graph = engin.graph()

    assert len(graph) == 3
